import os
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import (
    APTPackageHelper,
    HostNetworkingHelper,
    PebbleHelper,
    SnapPackageHelper,
    SystemdHelper,
)
from hotsos.core import plugintools

SERVICES = [r"etcd\S*",
            r"calico\S*",
            r"flannel\S*",
            r"containerd\S*",
            r"dockerd\S*",
            r"kubelet\S*",
            r"kube-\S*",
            r"microk8s\S*",
            ]

# Packages that only exist in a K8s deployment
K8S_PACKAGES = [r'cdk-addons',
                r'helm',
                r'kubernetes-[\S]+',
                r'kube-[\S]+',
                r'kubectl',
                r'kubelet',
                r'kubeadm',
                r'kubefed',
                r'microk8s'
                ]
# Packages that are used in a K8s deployment
K8S_PACKAGE_DEPS = [r'charm[\S]+',
                    r'docker',
                    r'go',
                    r'vault',
                    r'etcd',
                    r'conjure-up',
                    ]
# Snap-only deps
K8S_PACKAGE_DEPS_SNAP = [r'core[0-9]*']


class KubernetesBase(object):

    @cached_property
    def flannel_ports(self):
        ports = []
        nethelp = HostNetworkingHelper()
        for port in nethelp.host_interfaces:
            if "flannel" in port.name:
                ports.append(port)

        return ports

    @cached_property
    def bind_interfaces(self):
        """
        Fetch interfaces used by Kubernetes.
        """
        return {'flannel': self.flannel_ports}

    @cached_property
    def pods(self):
        pods = []
        pods_path = os.path.join(HotSOSConfig.data_root,
                                 "var/log/pods")
        if os.path.exists(pods_path):
            for pod in os.listdir(pods_path):
                pods.append(pod)

        return sorted(pods)

    @cached_property
    def containers(self):
        containers = []
        containers_path = os.path.join(HotSOSConfig.data_root,
                                       "var/log/containers")
        if os.path.exists(containers_path):
            for pod in os.listdir(containers_path):
                pod = pod.partition('.log')[0]
                containers.append(pod)

        return sorted(containers)


class KubernetesChecksBase(KubernetesBase, plugintools.PluginPartBase):
    plugin_name = 'kubernetes'
    plugin_root_index = 8

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        deps = K8S_PACKAGE_DEPS
        # Deployments can use snap or apt versions of packages so we check both
        self.apt = APTPackageHelper(core_pkgs=K8S_PACKAGES, other_pkgs=deps)
        snap_deps = deps + K8S_PACKAGE_DEPS_SNAP
        self.snaps = SnapPackageHelper(core_snaps=K8S_PACKAGES,
                                       other_snaps=snap_deps)
        self.pebble = PebbleHelper(service_exprs=SERVICES)
        self.systemd = SystemdHelper(service_exprs=SERVICES)

    @property
    def plugin_runnable(self):
        if self.apt.core or self.snaps.core:
            return True

        return False



from hotsos.core.plugins.maas import MAASChecksBase


class MAASSummary(MAASChecksBase):
    summary_part_index = 0

    def __0_summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

    def __1_summary_dpkg(self):
        if self.apt.core:
            return self.apt.all_formatted

    def __2_summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted



import os
import re

import yaml
from jinja2 import FileSystemLoader, Environment
from hotsos.core.config import HotSOSConfig
from hotsos.core.issues import IssuesManager
from hotsos.core.log import log
from hotsos.core.ycheck.scenarios import YScenarioChecker

PLUGINS = {}
PLUGIN_RUN_ORDER = []


class SummaryOffsetConflict(Exception):
    pass


class PluginRegistryMeta(type):
    """
    Each plugin that wants to add output to the finally summary must register
    itself by implementing one or more classes that have this class as the
    metaclass at some point in their resolution chain and:

        * class attribute 'plugin_root_index' which should be set once per
          plugin (usually in a high level base class) and defines the order in
          which plugins will be run.

        * class attribute 'summary_part_index' set to the integer index it
          should appear in the summary entry for that plugin. These indexes
          start at 0 and are unique to a plugin.
    """

    def __init__(cls, _name, _mro, members):
        name = members.get('plugin_name')
        if name:
            PLUGINS[name] = []
            PLUGIN_RUN_ORDER.append((name, members['plugin_root_index']))

        for subcls in cls.__mro__:
            index_key = 'summary_part_index'
            if hasattr(subcls, index_key):
                index = subcls.summary_part_index
                existing = [e[index_key] for e in PLUGINS[cls.plugin_name]]
                if index in existing:
                    raise SummaryOffsetConflict("plugin {} has index "
                                                "conflict {}".format(name,
                                                                     index))

                PLUGINS[cls.plugin_name].append({index_key: index,
                                                 'runner': subcls})


class HOTSOSDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)

    def represent_dict_preserve_order(self, data):
        return self.represent_dict(data.items())


def get_plugins_sorted():
    """ Return list of plugin names in the order they are to be run. """
    return [e[0] for e in sorted(PLUGIN_RUN_ORDER, key=lambda e: e[1])]


def yaml_dump(data):
    """
    This is our version of yaml.dump but ensuring the format/style that we
    want.
    """
    HOTSOSDumper.add_representer(
        dict,
        HOTSOSDumper.represent_dict_preserve_order)
    return yaml.dump(data, Dumper=HOTSOSDumper,
                     default_flow_style=False).rstrip("\n")


class OutputFormatterBase(object):

    def render(self, context, template):
        # jinja 2.10.x really needs this to be a str and e.g. not a PosixPath
        templates_dir = str(HotSOSConfig.templates_path)
        if not os.path.isdir(templates_dir):
            raise Exception("jinja templates directory not found: '{}'".
                            format(templates_dir))

        env = Environment(loader=FileSystemLoader(templates_dir))
        template = env.get_template(template)
        return template.render(context)


class HTMLFormatter(OutputFormatterBase):
    """
    Format the summary as html.

    Ref: https://iamkate.com/code/tree-views/
    """

    def __init__(self, hostname, max_level=2):
        """
        @param max_level: The HTML will be collapsible up to max_level
        """
        self.hostname = hostname
        self.max_level = max_level

    @property
    def header(self):
        return self.render({'hostname': self.hostname}, 'header.html')

    @property
    def footer(self):
        with open(os.path.join(HotSOSConfig.templates_path,
                               'footer.html')) as fd:
            return fd.read()

    def _expand_list(self, data, level):
        context = {'list_elements': [], 'indent': '    '}
        for item in data:
            context['list_elements'].append(self._expand(item, level))

        return self.render(context, 'content_list.html')

    def _expand_dict(self, data, level):
        context = {'dict_elements': {},
                   'indent': '    ',
                   'collapsible': level < self.max_level}
        if level == 1:
            context['class'] = 'class="tree"'

        for key, value in data.items():
            context['dict_elements'][key] = self._expand(value, level + 1)

        return self.render(context, 'content_dict.html')

    def _expand(self, data, level):
        """Expand the data object.

        @param data: The data object. This can be a dict, list, or a flat type
                     such as int or str.
        @param level: The current header level
        @return: A string expansion formatted in HTML of the data object.
        """
        if isinstance(data, dict):
            return self._expand_dict(data, level)
        if isinstance(data, list):
            return self._expand_list(data, level)

        return data

    def dump(self, data):
        """Convert the data (dict) into an html document.

        @param dist: The data
        @return: the html document as a string.
        """
        content = self._expand(data, 1)
        return self.header + content + self.footer


class MarkdownFormatter(OutputFormatterBase):

    def _expand_dict(self, data, level):
        level_prefix = ''.join(['#' for i in range(level)])
        markdown_output = ''
        for key in data:
            markdown_output += f'\n{level_prefix} {key}\n'
            markdown_output += self._expand(data[key], level + 1)

        return markdown_output

    def _expand_list(self, data):
        markdown_output = '\n%s' % ''.join(f'- {item}\n' for item in data)

        return markdown_output

    def _expand(self, data, level):
        """Expand the data object.

        @param data: The data object. This can be a dict, list, or a flat type
                     such as int or str.
        @param level: The current header level
        @return: string expansion formatted in markdown of the data object.
        """
        if isinstance(data, dict):
            return self._expand_dict(data, level)
        if isinstance(data, list):
            return self._expand_list(data)

        return f'\n{data}\n'

    def dump(self, data):
        """Convert the data (dict) into a markdown document.

        @param data: dict:- The data
        @return: str: The markdown document
        """
        markdown = '# hotsos summary\n' + self._expand(data, 2)
        return markdown.rstrip('\n')


class ApplicationBase(object, metaclass=PluginRegistryMeta):

    @property
    def bind_interfaces(self):
        """Implement this method to return a dict of network interfaces used
        by this application.
        """
        raise NotImplementedError


class SummaryEntry(object):

    def __init__(self, data, index):
        self.data = data
        self.index = index


class PartManager(object):

    def save(self, data, index):
        """
        Save part output yaml in temporary location. These are collected and
        aggregated at the end of the plugin run.
        """
        HOTSOSDumper.add_representer(
            dict,
            HOTSOSDumper.represent_dict_preserve_order)
        out = yaml.dump(data, Dumper=HOTSOSDumper,
                        default_flow_style=False).rstrip("\n")

        part_path = os.path.join(HotSOSConfig.plugin_tmp_dir,
                                 "{}.{}.part.yaml".
                                 format(HotSOSConfig.plugin_name,
                                        HotSOSConfig.part_name))

        # don't clobber
        if os.path.exists(part_path):
            newpath = part_path
            i = 0
            while os.path.exists(newpath):
                i += 1
                newpath = "{}.{}".format(part_path, i)

            part_path = newpath

        with open(part_path, 'w') as fd:
            fd.write(out)

        self.add_to_index(index, part_path)

    @property
    def indexes(self):
        path = os.path.join(HotSOSConfig.plugin_tmp_dir, "index.yaml")
        if os.path.exists(path):
            with open(path) as fd:
                return yaml.safe_load(fd.read()) or {}

        return {}

    def add_to_index(self, index, part):
        indexes = self.indexes
        path = os.path.join(HotSOSConfig.plugin_tmp_dir, "index.yaml")
        with open(path, 'w') as fd:
            if index in indexes:
                indexes[index].append(part)
            else:
                indexes[index] = [part]

            fd.write(yaml.dump(indexes))

    def meld_part_output(self, data, existing):
        """
        Don't allow root level keys to be clobbered, instead just
        update them. This assumes that part subkeys will be unique.
        """
        remove_keys = []
        for key in data:
            if key in existing:
                if isinstance(existing[key], dict):
                    existing[key].update(data[key])
                    remove_keys.append(key)

        if remove_keys:
            for key in remove_keys:
                del data[key]

        existing.update(data)

    def all(self):
        if not self.indexes:
            return

        parts = {}
        for index in sorted(self.indexes):
            for part in self.indexes[index]:
                with open(part) as fd:
                    part_yaml = yaml.safe_load(fd)

                    # Don't allow root level keys to be clobbered, instead just
                    # update them. This assumes that part subkeys will be
                    # unique.
                    self.meld_part_output(part_yaml, parts)

        return {HotSOSConfig.plugin_name: parts}


class PluginPartBase(ApplicationBase):
    # max per part before overlap
    PLUGIN_PART_INDEX_MAX = 500

    def __init__(self, *args, **kwargs):
        plugin_tmp = HotSOSConfig.plugin_tmp_dir
        if not plugin_tmp or not os.path.isdir(plugin_tmp):
            raise Exception("plugin plugin_tmp_dir not initialised - exiting")

        super().__init__(*args, **kwargs)

    @property
    def plugin_runnable(self):
        """
        Must be implemented by all plugins to define at runtime whether they
        should run.
        """
        raise NotImplementedError

    @property
    def summary_subkey(self):
        """
        This can be optionally implemented in order to have all output of the
        current part placed under the keyname provided by this property i.e.
        <plugin_name>: <subkey>: ...
        """
        return None

    @property
    def summary(self):
        """
        This can be optionally implemented as a way of pushing all results into
        the summary root for the current plugin part.

        The alternative is to implement methods called __summary_<key> such
        that the returned value is placed under <key>.
        """
        return None

    @property
    def output(self):
        _output = {}
        if self.summary:
            for key, data in self.summary.items():
                _output[key] = SummaryEntry(data, 0)

        for m in dir(self.__class__):
            cls = self.__class__.__name__
            ret = re.match(r'_{}__(\d+)_summary_(\S+)'.format(cls), m)
            if ret:
                index = int(ret.group(1))
                key = ret.group(2)
            else:
                ret = re.match(r'_{}__summary_(\S+)'.format(cls), m)
                if ret:
                    log.info("summary sub entry with no index: %s", m)
                    key = ret.group(1)
                    index = 0

            if not ret:
                continue

            out = getattr(self, m)()
            if not out:
                continue

            key = key.replace('_', '-')
            _output[key] = SummaryEntry(out, index)

        return _output

    @property
    def raw_output(self):
        out = self.output
        if out:
            return {key: entry.data for key, entry in out.items()}


class PluginRunner(object):

    def __init__(self, plugin):
        self.parts = PLUGINS[plugin]

    def run(self):
        part_mgr = PartManager()
        failed_parts = []
        # The following are executed as part of each plugin run (but not last).
        ALWAYS_RUN = {'auto_scenario_check': YScenarioChecker}
        for name, always_parts in ALWAYS_RUN.items():
            # update current env to reflect actual part being run
            HotSOSConfig.part_name = name
            try:
                always_parts().load_and_run()
            except Exception as exc:
                failed_parts.append(name)
                log.exception("part '%s' raised exception: %s", name, exc)

            # NOTE: we don't expect these parts to produce any output
            # for the summary so we wont check for it (they only raise
            # issues and bugs which are handled independently).

        for part_info in self.parts:
            # update current env to reflect actual part being run
            runner = part_info['runner']
            name = runner.__name__
            HotSOSConfig.part_name = name
            inst = runner()
            # Only run plugin if it declares itself runnable.
            if not HotSOSConfig.force_mode and not inst.plugin_runnable:
                log.debug("%s.%s.%s not runnable - skipping",
                          HotSOSConfig.plugin_name, name, runner.__name__)
                continue

            log.debug("running %s.%s.%s",
                      HotSOSConfig.plugin_name, name, runner.__name__)
            try:
                # NOTE: since all parts are expected to be implementations
                # of PluginPartBase we expect them to always define an
                # output property.
                output = inst.output
                subkey = inst.summary_subkey
            except Exception as exc:
                failed_parts.append(name)
                log.exception("part '%s' raised exception: %s", name, exc)
                output = None

            if output:
                for key, entry in output.items():
                    out = {key: entry.data}
                    if subkey:
                        out = {subkey: out}

                    part_max = PluginPartBase.PLUGIN_PART_INDEX_MAX
                    part_index = part_info['summary_part_index']
                    index = (part_index * part_max) + entry.index
                    part_mgr.save(out, index=index)

        if failed_parts:
            # always put these at the top
            part_mgr.save({'failed-parts': failed_parts}, index=0)

        imgr = IssuesManager()
        bugs = imgr.load_bugs()
        raised_issues = imgr.load_issues()
        summary_end_index = PluginPartBase.PLUGIN_PART_INDEX_MAX ** 2

        # Add detected known_bugs and raised issues to end summary.
        if bugs:
            part_mgr.save(bugs, index=summary_end_index)

        # Add raised issues to summary.
        if raised_issues:
            part_mgr.save(raised_issues, index=summary_end_index)

        return part_mgr.all()



from .cli import (  # noqa: F403,F401
    CLIHelper,
    CLIHelperFile,
)
from .config import (  # noqa: F403,F401
    ConfigBase,
    SectionalConfigBase,
)
from .network import (  # noqa: F403,F401
    NetworkPort,
    HostNetworkingHelper,
)
from .packaging import (  # noqa: F403,F401
    DPKGVersion,
    APTPackageHelper,
    DockerImageHelper,
    SnapPackageHelper,
)
from .pebble import (  # noqa: F403,F401 pylint: disable=cyclic-import
    PebbleHelper,
)
from .ssl import (  # noqa: F403,F401
    SSLCertificate,
    SSLCertificatesHelper,
)
from .systemd import (  # noqa: F403,F401 pylint: disable=cyclic-import
    SystemdHelper,
)
from .uptime import (  # noqa: F403,F401 pylint: disable=cyclic-import
    UptimeHelper,
)
from .sysctl import (  # noqa: F403,F401
    SYSCtlFactory,
    SYSCtlConfHelper,
)
from .apparmor import (  # noqa: F403,F401
    AAProfileFactory,
    ApparmorHelper,
)



from hotsos.core.plugins.sosreport import SOSReportChecksBase


class SOSReportSummary(SOSReportChecksBase):
    summary_part_index = 0

    def __0_summary_version(self):
        if self.version is not None:
            return self.version

    def __1_summary_dpkg(self):
        if self.apt.core:
            return self.apt.all_formatted

    def __2_summary_plugin_timeouts(self):
        if self.timed_out_plugins:
            return self.timed_out_plugins



from . import (  # noqa: F401
    summary,
    service_features,
    service_network_checks,
    vm_info,
    nova_external_events,
)
from .agent import (  # noqa: F403,F401
    events,
    exceptions,
)



import os
import re
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.plugins.openstack.openstack import (
    OpenstackConfig,
    OSTServiceBase,
)

# See https://github.com/openstack/neutron-lib/blob/master/neutron_lib/constants.py#L346  # noqa, pylint: disable=C0301
IP_HEADER_BYTES = 20
GRE_HEADER_BYTES = 22
VXLAN_HEADER_BYTES = 30


class NeutronBase(OSTServiceBase):

    def __init__(self, *args, **kwargs):
        super().__init__('neutron', *args, **kwargs)
        self.neutron_ovs_config = self.project.config['openvswitch-agent']

    @cached_property
    def bind_interfaces(self):
        """
        Fetch interfaces used by Openstack Neutron. Returned dict is keyed by
        config key used to identify interface.
        """
        local_ip = self.neutron_ovs_config.get('local_ip')

        interfaces = {}
        if not any([local_ip]):
            return interfaces

        if local_ip:
            port = self.nethelp.get_interface_with_addr(local_ip)
            # NOTE: local_ip can be an address or fqdn, we currently only
            # support searching by address.
            if port:
                interfaces.update({'local_ip': port})

        return interfaces


class ServiceChecks(object):

    @cached_property
    def ovs_cleanup_run_manually(self):
        """ Allow one run on node boot/reboot but not after. """
        run_manually = False
        start_count = 0
        cli = CLIHelper()
        cexpr = re.compile(r"Started OpenStack Neutron OVS cleanup.")
        for line in cli.journalctl(unit="neutron-ovs-cleanup"):
            if re.compile("-- Reboot --").match(line):
                # reset after reboot
                run_manually = False
                start_count = 0
            elif cexpr.search(line):
                if start_count:
                    run_manually = True

                start_count += 1

        return run_manually


class NeutronRouter(object):

    def __init__(self, uuid, ha_state):
        self.uuid = uuid
        self.ha_state = ha_state
        self.vr_id = None


class NeutronHAInfo(object):

    def __init__(self):
        self.vr_id = None

    @cached_property
    def state_path(self):
        ha_confs = 'var/lib/neutron/ha_confs'
        return os.path.join(HotSOSConfig.data_root, ha_confs)

    @cached_property
    def ha_routers(self):
        if not os.path.exists(self.state_path):
            return []

        _routers = []
        for entry in os.listdir(self.state_path):
            entry = os.path.join(self.state_path, entry)
            if not os.path.isdir(entry):
                # if its not a directory then it is probably not a live router
                # so we ignore it.
                continue

            state_path = os.path.join(entry, 'state')
            if not os.path.exists(state_path):
                continue

            with open(state_path) as fd:
                uuid = os.path.basename(entry)
                state = fd.read().strip()
                router = NeutronRouter(uuid, state)

            keepalived_conf_path = os.path.join(entry, 'keepalived.conf')
            if os.path.isfile(keepalived_conf_path):
                with open(keepalived_conf_path) as fd:
                    for line in fd:
                        expr = r'.+ virtual_router_id (\d+)'
                        ret = re.compile(expr).search(line)
                        if ret:
                            router.vr_id = ret.group(1)

            _routers.append(router)

        return _routers

    def find_router_with_vr_id(self, vr_id):
        for r in self.ha_routers:
            if r.vr_id == vr_id:
                return r


class Config(FactoryBase):

    def __getattr__(self, path):
        path = os.path.join(HotSOSConfig.data_root, 'etc/neutron', path)
        return OpenstackConfig(path)



import statistics
from datetime import datetime


class EventCollection(object):
    """Used to collect events found in logfiles. Events are defined as having
    identifiable start and end points containing timestamp information such
    that we can calculate their duration.

    Sequences can span multiple files and are identified by an "event_id" which
    may occur more than once. As such we always try to match using the closest
    available marker by date.
    """
    def __init__(self):
        self._events = {}

    def most_recent(self, items):
        """ For an event id that has been re-used, find the most recent one.

        This means that when we calculate stats on events found we will include
        only the most recent instance of an event with a given id.
        """
        return sorted(items, key=lambda e: e["end"], reverse=True)[0]

    @property
    def complete_events(self):
        """ Complete events are ones for which a duration has been
        calculated which implies their start and has been identified. """
        complete = {}
        for event_id, info in self._events.items():
            for item in info.get("heads", []):
                if "duration" in item:
                    if event_id not in complete:
                        complete[event_id] = []

                    complete[event_id].append(item)

            if event_id in complete:
                complete[event_id] = self.most_recent(complete[event_id])

        return complete

    @property
    def incomplete_events(self):
        incomplete = {}
        for event_id, info in self._events.items():
            for item in info.get("heads", []):
                if "duration" not in item:
                    if event_id not in incomplete:
                        incomplete[event_id] = []

                    incomplete[event_id].append(item)

        return incomplete

    def find_most_recent_start(self, event_id, end_ts):
        """
        For a given event end marker, find the most recent start marker.
        """
        last = {}
        for item in self._events[event_id].get("heads", []):
            start_ts = item["start"]
            if start_ts <= end_ts:
                if not last or start_ts > last["start"]:
                    last = item

        return last

    def add_event_end(self, event_id, end_ts):
        """
        Add an event termination marker with given event id and timestamp.

        We support non-unique event ids in the case that e.g. they get rotated
        by storing as a list of their corresponding end/tail timestamps.

        @param event_id: id used to identify an event iteration. This is
                         usually a sequence id but could be anything.
        @param end_ts: timestamp of the end of an event.
        """
        if event_id not in self._events:
            self._events[event_id] = {}

        if "tails" not in self._events[event_id]:
            self._events[event_id]["tails"] = [end_ts]
        else:
            self._events[event_id]["tails"].append(end_ts)

    def add_event_start(self, event_id, start_ts, metadata=None,
                        metadata_key=None):
        """
        Add an event start marker with given event id and timestamp.

        We support non-unique event ids in the case that e.g. they get rotated
        by storing as a list of their corresponding start timestamps. If an
        event includes some metadata that we want to collect we store it as
        part of the event start/head.

        @param event_id: id used to identify an event iteration. This is
                         usually a sequence id but could be anything.
        @param start_ts: timestamp of the start of an event.
        @param metadata: an optional field that can be extracted from the
                         search result.
        @param metadata_key: a custom key name used to store the metadata in
                             results.
        """
        event_info = {"start": start_ts}
        if metadata:
            if not metadata_key:
                metadata_key = "metadata"

            event_info[metadata_key] = metadata

        if event_id not in self._events:
            self._events[event_id] = {}

        if "heads" not in self._events[event_id]:
            self._events[event_id]["heads"] = [event_info]
        else:
            self._events[event_id]["heads"].append(event_info)

    def calculate_event_deltas(self):
        """
        Once we have collected all complete events i.e. ones that have a start
        and end, we can calculate their elapsed time. We add this to the start
        item.

        Since it is possible for events to be incomplete i.e. not have a start
        or end, we ensure to account for this by matching ends with their
        most recent start.
        """
        for event, info in self._events.items():
            _prev_start = None
            for end_ts in info.get("tails", []):
                start_item = self.find_most_recent_start(event, end_ts)
                if not start_item:
                    # incomplete event
                    continue

                new_start = start_item["start"]
                if _prev_start is None:
                    _prev_start = new_start
                elif _prev_start == new_start:
                    # If we have already closed a start/end loop, ignore any
                    # further endings
                    break

                etime = end_ts - new_start
                duration = round(float(etime.total_seconds()), 2)
                start_item["duration"] = duration
                start_item["end"] = end_ts


class SearchResultIndices(object):
    def __init__(self, day_idx=1, secs_idx=2, event_id_idx=3,
                 metadata_idx=None, metadata_key=None):
        """
        This is used to know where to find required information within a
        SearchResult. The indexes refer to python.re groups.

        The minimum required information that a result must contain is day,
        secs and event_id. Results will be referred to using whatever event_id
        is set to.
        """
        self.day = day_idx
        self.secs = secs_idx
        self.event_id = event_id_idx
        self.metadata = metadata_idx
        self.metadata_key = metadata_key


class LogEventStats(object):
    """Used to identify events within logs whereby a event has a start and end
    point. It can thenbe implemented by other classes to perform further
    analysis on event data.

    This class supports overlapping events e.g. for scenarios where logs
    are generated by parallal tasks.
    """

    def __init__(self, results, results_tag_prefix, custom_idxs=None):
        """
        @param results: FileSearcher results. This will be searched using
                        <results_tag_prefix>-start and <results_tag_prefix>-end
        @param results_tag_prefix: prefix of tag used for search results for
                                   events start and end.
        @param custom_idxs: optionally provide custom SearchResultIndices.
        """
        self.data = EventCollection()
        self.results = results
        self.results_tag_prefix = results_tag_prefix
        if custom_idxs:
            log_seq_idxs = custom_idxs
        else:
            log_seq_idxs = SearchResultIndices()

        self.log_seq_idxs = log_seq_idxs

    def run(self):
        """ Collect event start markers and end markers then attempt to link
        them to form complete events thus allowing us to calculate their
        duration.
        """
        seq_idxs = self.log_seq_idxs

        end_tag = "{}-end".format(self.results_tag_prefix)
        for result in self.results.find_by_tag(end_tag):
            day = result.get(seq_idxs.day)
            secs = result.get(seq_idxs.secs)
            end = "{} {}".format(day, secs)
            end = datetime.strptime(end, "%Y-%m-%d %H:%M:%S.%f")
            self.data.add_event_end(result.get(seq_idxs.event_id), end)

        start_tag = "{}-start".format(self.results_tag_prefix)
        for result in self.results.find_by_tag(start_tag):
            day = result.get(seq_idxs.day)
            secs = result.get(seq_idxs.secs)
            start = "{} {}".format(day, secs)
            start = datetime.strptime(start, "%Y-%m-%d %H:%M:%S.%f")
            metadata = result.get(seq_idxs.metadata)
            meta_key = seq_idxs.metadata_key
            event_id = result.get(seq_idxs.event_id)
            self.data.add_event_start(event_id, start, metadata=metadata,
                                      metadata_key=meta_key)

        self.data.calculate_event_deltas()

    def get_top_n_events_sorted(self, maximum, reverse=True):
        """
        Find events with the longest duration limited to max results.

        @param max: integer number of events to include.
        @param reverse: defaults to True for longest durations. Set to False to
        fetch events with shortest duration.
        @return: dictionary of results.
        """
        count = 0
        top_n = {}
        top_n_sorted = {}

        for event_id, item in sorted(self.data.complete_events.items(),
                                     key=lambda e: e[1]["duration"],
                                     reverse=reverse):
            if count >= maximum:
                break

            count += 1
            top_n[event_id] = item

        for event_id, item in sorted(top_n.items(),
                                     key=lambda x: x[1]["start"],
                                     reverse=reverse):
            top_n_sorted[event_id] = {"start": str(item["start"]),
                                      "end": str(item["end"]),
                                      "duration": item["duration"]}

            # include metadata in results if it is available.
            metadata_key = self.log_seq_idxs.metadata_key
            if item.get(metadata_key):
                top_n_sorted[event_id][metadata_key] = item[metadata_key]

        return top_n_sorted

    def get_event_stats(self):
        """ Return common statistics on the dataset of events. """
        events = self.data.complete_events
        if not events:
            return

        events = [s["duration"] for s in events.values()]
        stats = {'min': round(min(events), 2),
                 'max': round(max(events), 2),
                 'stdev': round(statistics.pstdev(events), 2),
                 'avg': round(statistics.mean(events), 2),
                 'samples': len(events)}

        if self.data.incomplete_events:
            stats['incomplete'] = len(self.data.incomplete_events.values())

        return stats
